"""Flax Linen backend for memax.

`memax.linen.gras` provides the Generalized Recurrent Algebraic Structure (GRAS) base module.
`memax.linen.groups` provides the algebraic structures (semigroups, groups, etc.) used in GRAS.
`memax.linen.set_actions` provides set action-based recurrent layers (slow RNNs).
`memax.linen.semigroups` provides semigroup-based recurrent layers (fast RNNs).
`memax.linen.scans` provides scan functions for recurrent updates.
"""