"""
memax: A library for memory-augmented neural networks.

We provide backends for both Equinox and Flax.

We utilize a Generalized Recurrent Algebraic Structure (GRAS) framework
to define and implement virtually any memory model we come across using a single interface.
Take a look at `memax.equinox.gras` or `memax.linen.gras` for more details.
"""